@@ -31,8 +31,11 @@ from mergin.client_pull import (
     download_project_finalize,
     download_project_is_running,
 )
-from mergin.client_pull import pull_project_async, pull_project_is_running, pull_project_finalize, pull_project_cancel
-from mergin.client_push import push_project_async, push_project_is_running, push_project_finalize, push_project_cancel
+from mergin.client_pull import pull_project_async, pull_project_is_running, pull_project_finalize, \
+    pull_project_cancel
+from mergin.client_push import push_project_async, push_project_is_running, push_project_finalize, \
+    push_project_cancel
+
 
 from pygeodiff import GeoDiff
 
@@ -134,7 +137,7 @@ def get_client(url=None, auth_token=None, username=None, password=None):
 
 
 def _print_unhandled_exception():
-    """ Outputs details of an unhandled exception that is being handled right now """
+    """Outputs details of an unhandled exception that is being handled right now"""
     click.secho("Unhandled exception!", fg="red")
     for line in traceback.format_exception(*sys.exc_info()):
         click.echo(line)
@@ -222,7 +225,7 @@ def create(ctx, project, public, from_dir):
 )
 @click.pass_context
 def list_projects(ctx, flag):
-    """List projects on the server"""
+    """List projects on the server."""
     filter_str = "(filter flag={})".format(flag) if flag is not None else "(all public)"
     click.echo("List of projects {}:".format(filter_str))
     mc = ctx.obj["client"]
@@ -243,7 +246,7 @@ def list_projects(ctx, flag):
 @click.option("--version", default=None, help="Version of project to download")
 @click.pass_context
 def download(ctx, project, directory, version):
-    """Download last version of mergin project"""
+    """Download last version of mergin project."""
     mc = ctx.obj["client"]
     if mc is None:
         return
@@ -269,6 +272,53 @@ def download(ctx, project, directory, version):
         _print_unhandled_exception()
 
 
+@cli.command()
+@click.argument("project")
+@click.argument("usernames", nargs=-1)
+@click.option("--permissions", help="permissions to be granted to project (reader, writer, owner)")
+@click.pass_context
+def share_add(ctx, project, usernames, permissions):
+    """Add permissions to [users] to project."""
+    mc = ctx.obj["client"]
+    if mc is None:
+        return
+    usernames = list(usernames)
+    mc.add_user_permissions_to_project(project, usernames, permissions)
+
+
+@cli.command()
+@click.argument("project")
+@click.argument("usernames", nargs=-1)
+@click.pass_context
+def share_remove(ctx, project, usernames):
+    """Remove [users] permissions from project."""
+    mc = ctx.obj["client"]
+    if mc is None:
+        return
+    usernames = list(usernames)
+    mc.remove_user_permissions_from_project(project, usernames)
+
+
+@cli.command()
+@click.argument("project")
+@click.pass_context
+def share(ctx, project):
+    """Fetch permissions to project."""
+    mc = ctx.obj["client"]
+    if mc is None:
+        return
+    access_list = mc.project_user_permissions(project)
+
+    for username in access_list.get("owners"):
+        click.echo("{:20}\t{:20}".format(username, "owner"))
+    for username in access_list.get("writers"):
+        if username not in access_list.get("owners"):
+            click.echo("{:20}\t{:20}".format(username, "writer"))
+    for username in access_list.get("readers"):
+        if username not in access_list.get("writers"):
+            click.echo("{:20}\t{:20}".format(username, "reader"))
+
+
 @cli.command()
 @click.argument("filepath")
 @click.argument("output")
@@ -311,7 +361,7 @@ def num_version(name):
 @cli.command()
 @click.pass_context
 def status(ctx):
-    """Show all changes in project files - upstream and local"""
+    """Show all changes in project files - upstream and local."""
     mc = ctx.obj["client"]
     if mc is None:
         return
@@ -326,6 +376,12 @@ def status(ctx):
     except Exception as e:
         _print_unhandled_exception()
         return
+
+    if mc.has_unfinished_pull():
+        click.secho("The previous pull has not finished completely: status "
+                    "of some files may be reported incorrectly. Use "
+                    "resolve_unfinished_pull command to try to fix that.", fg="yellow")
+
     click.secho("### Server changes:", fg="magenta")
     pretty_diff(pull_changes)
     click.secho("### Local changes:", fg="magenta")
@@ -337,7 +393,7 @@ def status(ctx):
 @cli.command()
 @click.pass_context
 def push(ctx):
-    """Upload local changes into Mergin repository"""
+    """Upload local changes into Mergin repository."""
     mc = ctx.obj["client"]
     if mc is None:
         return
@@ -369,7 +425,7 @@ def push(ctx):
 @cli.command()
 @click.pass_context
 def pull(ctx):
-    """Fetch changes from Mergin repository"""
+    """Fetch changes from Mergin repository."""
     mc = ctx.obj["client"]
     if mc is None:
         return
@@ -404,7 +460,7 @@ def pull(ctx):
 @click.argument("version")
 @click.pass_context
 def show_version(ctx, version):
-    """ Displays information about a single version of a project. `version` is 'v1', 'v2', etc. """
+    """Displays information about a single version of a project. `version` is 'v1', 'v2', etc."""
     mc = ctx.obj["client"]
     if mc is None:
         return
@@ -423,7 +479,7 @@ def show_version(ctx, version):
 @click.argument("path")
 @click.pass_context
 def show_file_history(ctx, path):
-    """ Displays information about a single version of a project """
+    """Displays information about a single version of a project."""
     mc = ctx.obj["client"]
     if mc is None:
         return
@@ -506,5 +562,24 @@ def remove(ctx, project):
         _print_unhandled_exception()
 
 
+@cli.command()
+@click.pass_context
+def resolve_unfinished_pull(ctx):
+    """Try to resolve unfinished pull."""
+    mc = ctx.obj["client"]
+    if mc is None:
+        return
+
+    try:
+        mc.resolve_unfinished_pull(os.getcwd())
+        click.echo("Unfinished pull successfully resolved")
+    except InvalidProject as e:
+        click.secho("Invalid project directory ({})".format(str(e)), fg="red")
+    except ClientError as e:
+        click.secho("Error: " + str(e), fg="red")
+    except Exception as e:
+        _print_unhandled_exception()
+
+
 if __name__ == "__main__":
     cli()
