@@ -1,6 +1,8 @@
 import logging
+import math
 import os
 import json
+import shutil
 import zlib
 import base64
 import urllib.parse
@@ -24,7 +26,7 @@ from .client_pull import (
 )
 from .client_pull import pull_project_async, pull_project_wait, pull_project_finalize
 from .client_push import push_project_async, push_project_wait, push_project_finalize
-from .utils import DateTimeEncoder, get_versions_with_file_changes
+from .utils import DateTimeEncoder, get_versions_with_file_changes, int_version
 from .version import __version__
 
 this_dir = os.path.dirname(os.path.realpath(__file__))
@@ -306,6 +308,23 @@ class MerginClient:
             return None  # not authenticated
         return self._user_info["username"]
 
+    def user_service(self):
+        """
+        Requests information about user from /user/service endpoint if such exists in self.url server.
+
+        Returns response from server as JSON dict or None if endpoint is not found
+        """
+
+        try:
+            response = self.get("/v1/user/service")
+        except ClientError as e:
+            self.log.debug("Unable to query for /user/service endpoint")
+            return
+
+        response = json.loads(response.read())
+
+        return response
+
     def create_project(self, project_name, is_public=False, namespace=None):
         """
         Create new project repository in user namespace on Mergin server.
@@ -455,17 +474,53 @@ class MerginClient:
         resp = self.get("/v1/project/{}".format(project_path), params)
         return json.load(resp)
 
-    def project_versions(self, project_path):
+    def project_versions(self, project_path, since=None, to=None):
         """
-        Get records of all project's versions (history).
+        Get records of project's versions (history) in ascending order.
+        If neither 'since' nor 'to' is specified it will return all versions.
 
         :param project_path: Project's full name (<namespace>/<name>)
         :type project_path: String
+        :param since: Version to track project history from
+        :type since: String
+        :param to: Version to track project history to
+        :type to: String
 
         :rtype: List[Dict]
         """
-        resp = self.get("/v1/project/version/{}".format(project_path))
-        return json.load(resp)
+        versions = []
+        per_page = 100  # server limit
+        num_since = int_version(since) if since else 1
+        num_to = int_version(to) if to else None  # we may not know yet
+        start_page = math.ceil(num_since / per_page)
+        if not num_to:
+            # let's get first page and count
+            params = {
+                "page": start_page,
+                "per_page": per_page,
+                "descending": False
+            }
+            resp = self.get("/v1/project/versions/paginated/{}".format(project_path), params)
+            resp_json = json.load(resp)
+            versions = resp_json["versions"]
+            num_to = resp_json['count']
+            latest_version = int_version(versions[-1]["name"])
+            if latest_version < num_to:
+                versions += self.project_versions(project_path, f"v{latest_version+1}", f"v{num_to}")
+        else:
+            end_page = math.ceil(num_to / per_page)
+            for page in range(start_page, end_page+1):
+                params = {
+                    "page": page,
+                    "per_page": per_page,
+                    "descending": False
+                }
+                resp = self.get("/v1/project/versions/paginated/{}".format(project_path), params)
+                versions += json.load(resp)["versions"]
+
+        # filter out versions not within range
+        filtered_versions = list(filter(lambda v: (num_since <= int_version(v["name"]) <= num_to), versions))
+        return filtered_versions
 
     def download_project(self, project_path, directory, version=None):
         """
@@ -519,6 +574,60 @@ class MerginClient:
             detail = f"Project path: {project_path}"
             raise ClientError(str(e), detail)
 
+    def add_user_permissions_to_project(self, project_path, usernames, permission_level):
+        """
+        Add specified permissions to specified users to project
+        :param project_path: project full name (<namespace>/<name>)
+        :param usernames: list of usernames to be granted specified permission level
+        :param permission_level: string (reader, writer, owner)
+        """
+        if permission_level not in ["owner", "reader", "writer"]:
+            raise ClientError("Unsupported permission level")
+
+        project_info = self.project_info(project_path)
+        access = project_info.get('access')
+        for name in usernames:
+            if permission_level == "owner":
+                access.get("ownersnames").append(name)
+            if permission_level == "writer" or permission_level == "owner":
+                access.get("writersnames").append(name)
+            if permission_level == "writer" or permission_level == "owner" or permission_level == "reader":
+                access.get("readersnames").append(name)
+        self.set_project_access(project_path, access)
+
+    def remove_user_permissions_from_project(self, project_path, usernames):
+        """
+       Removes specified users from project
+       :param project_path: project full name (<namespace>/<name>)
+       :param usernames: list of usernames to be granted specified permission level
+       """
+        project_info = self.project_info(project_path)
+        access = project_info.get('access')
+        for name in usernames:
+            if name in access.get("ownersnames"):
+                access.get("ownersnames").remove(name)
+            if name in access.get("writersnames"):
+                access.get("writersnames").remove(name)
+            if name in access.get("readersnames"):
+                access.get("readersnames").remove(name)
+        self.set_project_access(project_path, access)
+
+    def project_user_permissions(self, project_path):
+        """
+       Returns permissions for project
+       :param project_path: project full name (<namespace>/<name>)
+       :return dict("owners": list(usernames),
+                    "writers": list(usernames),
+                    "readers": list(usernames))
+       """
+        project_info = self.project_info(project_path)
+        access = project_info.get('access')
+        result = {}
+        result["owners"] = access.get("ownersnames")
+        result["writers"] = access.get("writersnames")
+        result["readers"] = access.get("readersnames")
+        return result
+
     def push_project(self, directory):
         """
         Upload local changes to the repository.
@@ -663,6 +772,93 @@ class MerginClient:
         :param version_to: ending project version tag for getting diff
         :type version_to: String
         """
-        job = download_diffs_async(self, project_dir, file_path, version_from, version_to)
+        mp = MerginProject(project_dir)
+        project_path = mp.metadata["name"]
+        file_history = self.project_file_history_info(project_path, file_path)
+        versions_to_fetch = get_versions_with_file_changes(
+            self, project_path, file_path, version_from=version_from, version_to=version_to, file_history=file_history
+        )
+        diffs = self.download_file_diffs(project_dir, file_path, versions_to_fetch[1:])
+        # concatenate diffs, if needed
+        output_dir = os.path.dirname(output_diff)
+        if len(diffs) >= 1:
+            os.makedirs(output_dir, exist_ok=True)
+            if len(diffs) > 1:
+                mp.geodiff.concat_changes(diffs, output_diff)
+            elif len(diffs) == 1:
+                shutil.copy(diffs[0], output_diff)
+
+    def download_file_diffs(self, project_dir, file_path, versions):
+        """ Download file diffs for specified versions.
+
+        :param project_dir: project local directory
+        :type project_dir: String
+        :param file_path: relative path of file to download in the project directory
+        :type file_path: String
+        :param versions: list of versions to download diffs for, for example ['v3', 'v5']
+        :type versions: List(String)
+        :returns: list of downloaded diffs (their actual locations on disk)
+        :rtype: List(String)
+        """
+        job = download_diffs_async(self, project_dir, file_path, versions)
         pull_project_wait(job)
-        download_diffs_finalize(job, output_diff)
+        diffs = download_diffs_finalize(job)
+        return diffs
+
+    def has_unfinished_pull(self, directory):
+        """
+        Test whether project has an unfinished pull.
+
+        Unfinished pull means that a previous pull_project() call has
+        failed in the final stage due to some files being in read-only
+        mode. When a project has unfinished pull, it has to be resolved
+        before allowing further pulls or pushes.
+
+        .. seealso:: self.resolve_unfinished_pull
+
+        :param directory: project's directory
+        :type directory: String
+        :returns: project has unfinished pull
+        rtype: Boolean
+        """
+        mp = MerginProject(directory)
+
+        return mp.has_unfinished_pull()
+
+    def resolve_unfinished_pull(self, directory):
+        """
+        Try to resolve unfinished pull.
+
+        Unfinished pull means that a previous pull_project() call has
+        failed in the final stage due to some files being in read-only
+        mode. When a project has unfinished pull, it has to be resolved
+        before allowing further pulls or pushes.
+
+        To resolving unfinihed pull includes creation of conflicted copy
+        and replacement of the original file by the new version from the
+        server.
+
+        .. seealso:: self.has_unfinished_pull
+
+        :param directory: project's directory
+        :type directory: str
+        :returns: files where conflicts were found
+        :rtype: list[str]
+        """
+        mp = MerginProject(directory)
+        conflicts = mp.resolve_unfinished_pull(self.username())
+        return conflicts
+
+    def has_writing_permissions(self, project_path):
+        """
+        Test whether user has writing permissions on the given project.
+        We rely on the "permissions" field here, as it provides more accurate
+        information and take into account namespace permissions too.
+
+        :param project_path: project's path on server in the form or namespace/project
+        :type directory: str
+        :returns: whether user has writing (upload) permission on the project
+        :rtype: bool
+        """
+        info = self.project_info(project_path)
+        return info["permissions"]["upload"]
