@@ -21,7 +21,7 @@ from .merginproject import MerginProject
 
 class UploadJob:
     """ Keeps all the important data about a pending upload job """
-    
+
     def __init__(self, project_path, changes, transaction_id, mp, mc, tmp_dir):
         self.project_path = project_path       # full project name ("username/projectname")
         self.changes = changes                 # dictionary of local changes to the project
@@ -46,16 +46,16 @@ class UploadJob:
 
 class UploadQueueItem:
     """ A single chunk of data that needs to be uploaded """
-    
+
     def __init__(self, file_path, size, transaction_id, chunk_id, chunk_index):
         self.file_path = file_path            # full path to the file
         self.size = size                      # size of the chunk in bytes
         self.chunk_id = chunk_id              # ID of the chunk within transaction
         self.chunk_index = chunk_index        # index (starting from zero) of the chunk within the file
         self.transaction_id = transaction_id  # ID of the transaction
-    
+
     def upload_blocking(self, mc, mp):
-        
+
         with open(self.file_path, 'rb') as file_handle:
             file_handle.seek(self.chunk_index * UPLOAD_CHUNK_SIZE)
             data = file_handle.read(UPLOAD_CHUNK_SIZE)
@@ -70,14 +70,20 @@ class UploadQueueItem:
             resp_dict = json.load(resp)
             mp.log.debug(f"Upload finished: {self.file_path}")
             if not (resp_dict['size'] == len(data) and resp_dict['checksum'] == checksum.hexdigest()):
-                mc.post("/v1/project/push/cancel/{}".format(self.transaction_id))
+                try:
+                    mc.post("/v1/project/push/cancel/{}".format(self.transaction_id))
+                except ClientError:
+                    pass
                 raise ClientError("Mismatch between uploaded file chunk {} and local one".format(self.chunk_id))
-        
+
 
 def push_project_async(mc, directory):
     """ Starts push of a project and returns pending upload job """
 
     mp = MerginProject(directory)
+    if mp.has_unfinished_pull():
+        raise ClientError("Project is in unfinished pull state. Please resolve unfinished pull and try again.")
+
     project_path = mp.metadata["name"]
     local_version = mp.metadata["version"]
 
@@ -95,7 +101,10 @@ def push_project_async(mc, directory):
     mp.log.info(f"got project info: local version {local_version} / server version {server_version}")
 
     username = mc.username()
-    if username not in server_info["access"]["writersnames"]:
+    # permissions field contains information about update, delete and upload privileges of the user
+    # on a specific project. This is more accurate information then "writernames" field, as it takes
+    # into account namespace privileges. So we have to check only "permissions", namely "upload" one
+    if not mc.has_writing_permissions(project_path):
         mp.log.error(f"--- push {project_path} - username {username} does not have write access")
         raise ClientError(f"You do not seem to have write access to the project (username '{username}')")
 
@@ -202,7 +211,7 @@ def push_project_async(mc, directory):
 
 def push_project_wait(job):
     """ blocks until all upload tasks are finished """
-    
+
     concurrent.futures.wait(job.futures)
 
 
@@ -263,9 +272,11 @@ def push_project_finalize(job):
             # if push finish fails, the transaction is not killed, so we
             # need to cancel it so it does not block further uploads
             job.mp.log.info("canceling the pending transaction...")
-            resp_cancel = job.mc.post("/v1/project/push/cancel/%s" % job.transaction_id)
-            server_resp_cancel = json.load(resp_cancel)
-            job.mp.log.info("cancel response: " + str(server_resp_cancel))
+            try:
+                resp_cancel = job.mc.post("/v1/project/push/cancel/%s" % job.transaction_id)
+                job.mp.log.info("cancel response: " + resp_cancel.msg)
+            except ClientError as err2:
+                job.mp.log.info("cancel response: " + str(err2))
             raise err
 
     job.mp.metadata = {
@@ -273,7 +284,12 @@ def push_project_finalize(job):
         'version': job.server_resp['version'],
         'files': job.server_resp["files"]
     }
-    job.mp.apply_push_changes(job.changes)
+    try:
+        job.mp.apply_push_changes(job.changes)
+    except Exception as e:
+        job.mp.log.error("Failed to apply push changes: " + str(e))
+        job.mp.log.info("--- push aborted")
+        raise ClientError("Failed to apply push changes: " + str(e))
 
     job.tmp_dir.cleanup()   # delete our temporary dir and all its content
 
@@ -293,7 +309,7 @@ def push_project_cancel(job):
     job.executor.shutdown(wait=True)
     try:
         resp_cancel = job.mc.post("/v1/project/push/cancel/%s" % job.transaction_id)
-        job.server_resp = json.load(resp_cancel)
+        job.server_resp = resp_cancel.msg
     except ClientError as err:
         job.mp.log.error("--- push cancelling failed! " + str(err))
         raise err
