@@ -6,11 +6,21 @@ import os
 import re
 import shutil
 import uuid
+import tempfile
 from datetime import datetime
 from dateutil.tz import tzlocal
 
 from .common import UPLOAD_CHUNK_SIZE, InvalidProject, ClientError
-from .utils import generate_checksum, move_file, int_version, find, do_sqlite_checkpoint
+from .utils import (
+    generate_checksum,
+    move_file,
+    int_version,
+    find,
+    do_sqlite_checkpoint,
+    unique_path_name,
+    conflicted_copy_file_name,
+    edit_conflict_file_name
+)
 
 
 this_dir = os.path.dirname(os.path.realpath(__file__))
@@ -21,7 +31,7 @@ this_dir = os.path.dirname(os.path.realpath(__file__))
 # python paths.
 try:
     from .deps import pygeodiff
-except ImportError:
+except (ImportError, ModuleNotFoundError):
     import pygeodiff
 
 
@@ -39,6 +49,13 @@ class MerginProject:
         if not os.path.exists(self.meta_dir):
             os.mkdir(self.meta_dir)
 
+        # location for files from unfinished pull
+        self.unfinished_pull_dir = os.path.join(self.meta_dir, 'unfinished_pull')
+
+        self.cache_dir = os.path.join(self.meta_dir, '.cache')
+        if not os.path.exists(self.cache_dir):
+            os.mkdir(self.cache_dir)
+
         self.setup_logging(directory)
 
         # make sure we can load correct pygeodiff
@@ -68,7 +85,7 @@ class MerginProject:
         if not self.log.handlers:
             # we only need to set the handler once
             # (otherwise we would get things logged multiple times as loggers are cached)
-            log_handler = logging.FileHandler(os.path.join(self.meta_dir, "client-log.txt"))
+            log_handler = logging.FileHandler(os.path.join(self.meta_dir, "client-log.txt"), encoding = 'utf-8')
             log_handler.setFormatter(logging.Formatter('%(asctime)s %(message)s'))
             self.log.addHandler(log_handler)
 
@@ -102,6 +119,18 @@ class MerginProject:
         """ Helper function to get absolute path of file in meta dir. """
         return self.fpath(file, self.meta_dir)
 
+    def fpath_unfinished_pull(self, file):
+        """ Helper function to get absolute path of file in unfinished_pull dir. """
+        return self.fpath(file, self.unfinished_pull_dir)
+
+    def fpath_cache(self, file, version=None):
+        """ Helper function to get absolute path of file in cache dir.
+        It can be either in root cache directory (.mergin/.cache/) or in some version's subfolder
+        """
+        if version:
+            return self.fpath(file, os.path.join(self.cache_dir, version))
+        return self.fpath(file, self.cache_dir)
+
     @property
     def metadata(self):
         if not os.path.exists(self.fpath_meta('mergin.json')):
@@ -235,7 +264,8 @@ class MerginProject:
 
         .. seealso:: self.compare_file_sets
 
-        :param server_files: list of server files' metadata (see also self.inspect_files())
+        :param server_files: list of server files' metadata with mandatory 'history' field (see also self.inspect_files(),
+        self.project_info(project_path, since=v1))
         :type server_files: list[dict]
         :returns: changes metadata for files to be pulled from server
         :rtype: dict
@@ -372,7 +402,7 @@ class MerginProject:
                     pass
         return changes
 
-    def apply_pull_changes(self, changes, temp_dir):
+    def apply_pull_changes(self, changes, temp_dir, user_name):
         """
         Apply changes pulled from server.
 
@@ -409,69 +439,17 @@ class MerginProject:
                 # 'src' here is server version of file and 'dest' is locally modified
                 if self.is_versioned_file(path) and k == 'updated':
                     if path in modified:
-                        self.log.info("updating file with rebase: " + path)
-                        server_diff = self.fpath(f'{path}-server_diff', temp_dir)  # diff between server file and local basefile
-                        local_diff = self.fpath(f'{path}-local_diff', temp_dir)
-
-                        # temporary backup of file pulled from server for recovery
-                        f_server_backup = self.fpath(f'{path}-server_backup', temp_dir)
-                        self.geodiff.make_copy_sqlite(src, f_server_backup)
-
-                        # create temp backup (ideally with geodiff) of locally modified file if needed later
-                        f_conflict_file = self.fpath(f'{path}-local_backup', temp_dir)
-                        try:
-                            self.geodiff.create_changeset(basefile, dest, local_diff)
-                            self.geodiff.make_copy_sqlite(basefile, f_conflict_file)
-                            self.geodiff.apply_changeset(f_conflict_file, local_diff)
-                        except (pygeodiff.GeoDiffLibError, pygeodiff.GeoDiffLibConflictError):
-                            self.log.info("backup of local file with geodiff failed - need to do hard copy")
-                            self.geodiff.make_copy_sqlite(dest, f_conflict_file)
-
-                        # in case there will be any conflicting operations found during rebase,
-                        # they will be stored in a JSON file - if there are no conflicts, the file
-                        # won't even be created
-                        rebase_conflicts = self.fpath(f'{path}_rebase_conflicts')
-
-                        # try to do rebase magic
-                        try:
-                            self.geodiff.create_changeset(basefile, src, server_diff)
-                            self.geodiff.rebase(basefile, src, dest, rebase_conflicts)
-                            # make sure basefile is in the same state as remote server file (for calc of push changes)
-                            self.geodiff.apply_changeset(basefile, server_diff)
-                            self.log.info("rebase successful!")
-                        except (pygeodiff.GeoDiffLibError, pygeodiff.GeoDiffLibConflictError) as err:
-                            self.log.warning("rebase failed! going to create conflict file")
-                            # it would not be possible to commit local changes, they need to end up in new conflict file
-                            self.geodiff.make_copy_sqlite(f_conflict_file, dest)
-                            conflict = self.backup_file(path)
+                        conflict = self.update_with_rebase(path, src, dest, basefile, temp_dir, user_name)
+                        if conflict:
                             conflicts.append(conflict)
-                            # original file synced with server
-                            self.geodiff.make_copy_sqlite(f_server_backup, basefile)
-                            self.geodiff.make_copy_sqlite(f_server_backup, dest)
                     else:
                         # The local file is not modified -> no rebase needed.
                         # We just apply the diff between our copy and server to both the local copy and its basefile
-                        self.log.info("updating file without rebase: " + path)
-                        try:
-                            server_diff = self.fpath(f'{path}-server_diff', temp_dir)  # diff between server file and local basefile
-                            # TODO: it could happen that basefile does not exist.
-                            # It was either never created (e.g. when pushing without geodiff)
-                            # or it was deleted by mistake(?) by the user. We should detect that
-                            # when starting pull and download it as well
-                            self.geodiff.create_changeset(basefile, src, server_diff)
-                            self.geodiff.apply_changeset(dest, server_diff)
-                            self.geodiff.apply_changeset(basefile, server_diff)
-                            self.log.info("update successful")
-                        except (pygeodiff.GeoDiffLibError, pygeodiff.GeoDiffLibConflictError):
-                            self.log.warning("update failed! going to copy file")
-                            # something bad happened and we have failed to patch our local files - this should not happen if there
-                            # wasn't a schema change or something similar that geodiff can't handle.
-                            self.geodiff.make_copy_sqlite(src, dest)
-                            self.geodiff.make_copy_sqlite(src, basefile)
+                        self.update_without_rebase(path, src, dest, basefile, temp_dir)
                 else:
                     # backup if needed
                     if path in modified and item['checksum'] != local_files_map[path]['checksum']:
-                        conflict = self.backup_file(path)
+                        conflict = self.create_conflicted_copy(path, user_name)
                         conflicts.append(conflict)
 
                     if k == 'removed':
@@ -491,6 +469,115 @@ class MerginProject:
 
         return conflicts
 
+    def update_with_rebase(self, path, src, dest, basefile, temp_dir, user_name):
+        """
+        Update a versioned file with rebase.
+
+        Try to peform automatic rebase, create conflict file if failed.pyt
+
+        .. seealso:: self.update_without_rebase
+
+        :param path: path to geodiff file
+        :type path: str
+        :param src: path to the server version of the file
+        :type src: str
+        :param dest: path to the local version of the file
+        :type dest: str
+        :param basefile: path to a file in meta dir
+        :type basefile: str
+        :param temp_dir: directory with downloaded files from server
+        :type temp_dir: str
+        :returns: path to conflict file if rebase fails, empty string on success
+        :rtype: str
+        """
+        self.log.info("updating file with rebase: " + path)
+
+        server_diff = self.fpath(f'{path}-server_diff', temp_dir)  # diff between server file and local basefile
+        local_diff = self.fpath(f'{path}-local_diff', temp_dir)
+
+        # temporary backup of file pulled from server for recovery
+        f_server_backup = self.fpath(f'{path}-server_backup', temp_dir)
+        self.geodiff.make_copy_sqlite(src, f_server_backup)
+
+        # create temp backup (ideally with geodiff) of locally modified file if needed later
+        f_conflict_file = self.fpath(f'{path}-local_backup', temp_dir)
+
+        try:
+            self.geodiff.create_changeset(basefile, dest, local_diff)
+            self.geodiff.make_copy_sqlite(basefile, f_conflict_file)
+            self.geodiff.apply_changeset(f_conflict_file, local_diff)
+        except (pygeodiff.GeoDiffLibError, pygeodiff.GeoDiffLibConflictError):
+            self.log.info("backup of local file with geodiff failed - need to do hard copy")
+            self.geodiff.make_copy_sqlite(dest, f_conflict_file)
+
+        # in case there will be any conflicting operations found during rebase,
+        # they will be stored in a JSON file - if there are no conflicts, the file
+        # won't even be created
+        rebase_conflicts = unique_path_name(
+                edit_conflict_file_name(self.fpath(path), user_name, int_version(self.metadata['version'])))
+
+        # try to do rebase magic
+        try:
+            self.geodiff.create_changeset(basefile, src, server_diff)
+            self.geodiff.rebase(basefile, src, dest, rebase_conflicts)
+            # make sure basefile is in the same state as remote server file (for calc of push changes)
+            self.geodiff.apply_changeset(basefile, server_diff)
+            self.log.info("rebase successful!")
+        except (pygeodiff.GeoDiffLibError, pygeodiff.GeoDiffLibConflictError) as err:
+            self.log.warning("rebase failed! going to create conflict file")
+            try:
+                # it would not be possible to commit local changes, they need to end up in new conflict file
+                self.geodiff.make_copy_sqlite(f_conflict_file, dest)
+                conflict = self.create_conflicted_copy(path, user_name)
+                # original file synced with server
+                self.geodiff.make_copy_sqlite(f_server_backup, basefile)
+                self.geodiff.make_copy_sqlite(f_server_backup, dest)
+                return conflict
+            except pygeodiff.GeoDiffLibError as err:
+                self.log.warning("creation of conflicted copy failed! going to create an unfinished pull")
+                f_server_unfinished = self.fpath_unfinished_pull(path)
+                self.geodiff.make_copy_sqlite(f_server_backup, f_server_unfinished)
+
+        return ''
+
+    def update_without_rebase(self, path, src, dest, basefile, temp_dir):
+        """
+        Update a versioned file without rebase.
+
+        Apply the diff between local copy and server to both the local
+        copy and its basefile.
+
+        .. seealso:: self.update_with_rebase
+
+        :param path: path to geodiff file
+        :type path: str
+        :param src: path to the server version of the file
+        :type src: str
+        :param dest: path to the local version of the file
+        :type dest: str
+        :param basefile: path to a file in meta dir
+        :type basefile: str
+        :param temp_dir: directory with downloaded files from server
+        :type temp_dir: str
+        """
+        self.log.info("updating file without rebase: " + path)
+        try:
+            server_diff = self.fpath(f'{path}-server_diff', temp_dir)  # diff between server file and local basefile
+            # TODO: it could happen that basefile does not exist.
+            # It was either never created (e.g. when pushing without geodiff)
+            # or it was deleted by mistake(?) by the user. We should detect that
+            # when starting pull and download it as well
+            self.geodiff.create_changeset(basefile, src, server_diff)
+            self.geodiff.apply_changeset(dest, server_diff)
+            self.geodiff.apply_changeset(basefile, server_diff)
+            self.log.info("update successful")
+        except (pygeodiff.GeoDiffLibError, pygeodiff.GeoDiffLibConflictError):
+            self.log.warning("update failed! going to copy file")
+            # something bad happened and we have failed to patch our local files - this should not happen if there
+            # wasn't a schema change or something similar that geodiff can't handle.
+            self.geodiff.make_copy_sqlite(src, dest)
+            self.geodiff.make_copy_sqlite(src, basefile)
+
     def apply_push_changes(self, changes):
         """
         For geodiff files update basefiles according to changes pushed to server.
@@ -526,23 +613,21 @@ class MerginProject:
                 else:
                     pass
 
-    def backup_file(self, file):
+    def create_conflicted_copy(self, file, user_name):
         """
-        Create backup file next to its origin.
+        Create conflicted copy file next to its origin.
 
         :param file: path of file in project
         :type file: str
-        :returns: path to backupfile
+        :returns: path to conflicted copy
         :rtype: str
         """
         src = self.fpath(file)
         if not os.path.exists(src):
             return
-        backup_path = self.fpath(f'{file}_conflict_copy')
-        index = 2
-        while os.path.exists(backup_path):
-            backup_path = self.fpath(f'{file}_conflict_copy{index}')
-            index += 1
+
+        backup_path = unique_path_name(conflicted_copy_file_name(self.fpath(file), user_name, int_version(self.metadata['version'])))
+
         if self.is_versioned_file(file):
             self.geodiff.make_copy_sqlite(src, backup_path)
         else:
@@ -573,3 +658,75 @@ class MerginProject:
                 error = str(e)
                 break
         return error
+
+    def has_unfinished_pull(self):
+        """ Check if there is an unfinished pull for this project.
+
+        Unfinished pull means that a previous pull_project() call has
+        failed in the final stage due to some files being in read-only
+        mode. When a project has unfinished pull, it has to be resolved
+        before allowing further pulls or pushes.
+
+        .. seealso:: self.resolve_unfinished_pull
+
+        :returns: whether there is an unfinished pull
+        :rtype: bool
+        """
+        return os.path.exists(self.unfinished_pull_dir)
+
+    def resolve_unfinished_pull(self, user_name):
+        """
+        Try to resolve unfinished pull.
+
+        Unfinished pull means that a previous pull_project() call has
+        failed in the final stage due to some files being in read-only
+        mode. When a project has unfinished pull, it has to be resolved
+        before allowing further pulls or pushes.
+
+        Resolving unfinihed pull means creation of the conflicted copy
+        and replacement of the original file by the new version from the
+        server.
+
+        .. seealso:: self.has_unfinished_pull
+
+        :param user_name: name of the user
+        :type user_name: str
+        :returns: files where conflicts were found
+        :rtype: list[str]
+        """
+        conflicts = []
+
+        if not self.has_unfinished_pull():
+            self.log.warning("no unfinished pulls found")
+            return
+
+        self.log.info("resolving unfinished pull")
+
+        temp_dir = tempfile.mkdtemp(prefix="mergin-py-client-")
+
+        for root, dirs, files in os.walk(self.unfinished_pull_dir):
+            for file_name in files:
+                src = os.path.join(root, file_name)
+                dest = self.fpath(file_name)
+                basefile = self.fpath_meta(file_name)
+
+                self.log.info("trying to resolve unfinished pull for: " + file_name)
+
+                # 'src' here is a server version of the file from unfinished
+                # pull and 'dest' is a local version of the same file.
+                # to resolve unfinished pull we create a conflicted copy and
+                # replace local file with the changes from the server.
+                try:
+                    # conflicted copy
+                    conflict = self.create_conflicted_copy(dest, user_name)
+                    conflicts.append(conflict)
+                    # original file synced with server
+                    self.geodiff.make_copy_sqlite(src, basefile)
+                    self.geodiff.make_copy_sqlite(src, dest)
+                except pygeodiff.GeoDiffLibError as err:
+                    self.log.error("unable to apply changes from previous unfinished pull!")
+                    raise ClientError("Unable to resolve unfinished pull!")
+
+        shutil.rmtree(self.unfinished_pull_dir)
+        self.log.info("unfinished pull resolved successfuly!")
+        return conflicts
